from collections import Counter
from unittest import mock

import pytest
from werkzeug.test import Client

import sentry_sdk
from sentry_sdk import capture_message
from sentry_sdk.integrations.wsgi import (
    SentryWsgiMiddleware,
    _ScopedResponse,
    get_request_url,
)


@pytest.fixture
def crashing_app():
    def app(environ, start_response):
        1 / 0

    return app


class IterableApp:
    def __init__(self, iterable):
        self.iterable = iterable

    def __call__(self, environ, start_response):
        return self.iterable


class ExitingIterable:
    def __init__(self, exc_func):
        self._exc_func = exc_func

    def __iter__(self):
        return self

    def __next__(self):
        raise self._exc_func()

    def next(self):
        return type(self).__next__(self)


def test_basic(sentry_init, crashing_app, capture_events):
    sentry_init(send_default_pii=True)
    app = SentryWsgiMiddleware(crashing_app)
    client = Client(app)
    events = capture_events()

    with pytest.raises(ZeroDivisionError):
        client.get("/")

    (event,) = events

    assert event["transaction"] == "generic WSGI request"

    assert event["request"] == {
        "env": {"SERVER_NAME": "localhost", "SERVER_PORT": "80"},
        "headers": {"Host": "localhost"},
        "method": "GET",
        "query_string": "",
        "url": "http://localhost/",
    }


@pytest.mark.parametrize("path_info", ("bark/", "/bark/"))
@pytest.mark.parametrize("script_name", ("woof/woof", "woof/woof/"))
def test_script_name_is_respected(
    sentry_init, crashing_app, capture_events, script_name, path_info
):
    sentry_init(send_default_pii=True)
    app = SentryWsgiMiddleware(crashing_app)
    client = Client(app)
    events = capture_events()

    with pytest.raises(ZeroDivisionError):
        # setting url with PATH_INFO: bark/, HTTP_HOST: dogs.are.great and SCRIPT_NAME: woof/woof/
        client.get(path_info, f"https://dogs.are.great/{script_name}")  # noqa: E231

    (event,) = events

    assert event["request"]["url"] == "https://dogs.are.great/woof/woof/bark/"


@pytest.fixture(params=[0, None])
def test_systemexit_zero_is_ignored(sentry_init, capture_events, request):
    zero_code = request.param
    sentry_init(send_default_pii=True)
    iterable = ExitingIterable(lambda: SystemExit(zero_code))
    app = SentryWsgiMiddleware(IterableApp(iterable))
    client = Client(app)
    events = capture_events()

    with pytest.raises(SystemExit):
        client.get("/")

    assert len(events) == 0


@pytest.fixture(params=["", "foo", 1, 2])
def test_systemexit_nonzero_is_captured(sentry_init, capture_events, request):
    nonzero_code = request.param
    sentry_init(send_default_pii=True)
    iterable = ExitingIterable(lambda: SystemExit(nonzero_code))
    app = SentryWsgiMiddleware(IterableApp(iterable))
    client = Client(app)
    events = capture_events()

    with pytest.raises(SystemExit):
        client.get("/")

    (event,) = events

    assert "exception" in event
    exc = event["exception"]["values"][-1]
    assert exc["type"] == "SystemExit"
    assert exc["value"] == nonzero_code
    assert event["level"] == "error"


def test_keyboard_interrupt_is_captured(sentry_init, capture_events):
    sentry_init(send_default_pii=True)
    iterable = ExitingIterable(lambda: KeyboardInterrupt())
    app = SentryWsgiMiddleware(IterableApp(iterable))
    client = Client(app)
    events = capture_events()

    with pytest.raises(KeyboardInterrupt):
        client.get("/")

    (event,) = events

    assert "exception" in event
    exc = event["exception"]["values"][-1]
    assert exc["type"] == "KeyboardInterrupt"
    assert exc["value"] == ""
    assert event["level"] == "error"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_with_error(
    sentry_init,
    crashing_app,
    capture_events,
    capture_items,
    DictionaryContaining,  # noqa:N803
    span_streaming,
):
    def dogpark(environ, start_response):
        raise ValueError("Fetch aborted. The ball was not returned.")

    sentry_init(
        send_default_pii=True,
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(dogpark)
    client = Client(app)

    if span_streaming:
        items = capture_items("event", "span")
    else:
        events = capture_events()

    with pytest.raises(ValueError):
        client.get("http://dogs.are.great/sit/stay/rollover/")

    sentry_sdk.flush()

    if span_streaming:
        assert len(items) == 2
        assert items[0].type == "event"
        assert items[1].type == "span"

        error_event = items[0].payload
        span_item = items[1].payload
    else:
        error_event, envelope = events

        assert error_event["transaction"] == "generic WSGI request"

    assert error_event["contexts"]["trace"]["op"] == "http.server"
    assert error_event["exception"]["values"][0]["type"] == "ValueError"
    assert error_event["exception"]["values"][0]["mechanism"]["type"] == "wsgi"
    assert error_event["exception"]["values"][0]["mechanism"]["handled"] is False
    assert (
        error_event["exception"]["values"][0]["value"]
        == "Fetch aborted. The ball was not returned."
    )

    if span_streaming:
        assert span_item["trace_id"] == error_event["contexts"]["trace"]["trace_id"]
        assert span_item["span_id"] == error_event["contexts"]["trace"]["span_id"]
        assert span_item["status"] == "error"
    else:
        assert envelope["type"] == "transaction"

        # event trace context is a subset of envelope trace context
        assert envelope["contexts"]["trace"] == DictionaryContaining(
            error_event["contexts"]["trace"]
        )
        assert envelope["contexts"]["trace"]["status"] == "internal_error"
        assert envelope["transaction"] == error_event["transaction"]
        assert envelope["request"] == error_event["request"]


@pytest.mark.parametrize("send_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_no_error(
    sentry_init,
    capture_events,
    capture_items,
    DictionaryContaining,  # noqa:N803
    span_streaming,
    send_pii,
):
    def dogpark(environ, start_response):
        start_response("200 OK", [])
        return ["Go get the ball! Good dog!"]

    sentry_init(
        send_default_pii=send_pii,
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(dogpark)
    client = Client(app)

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client.get("/dogs/are/great?toy=tennisball")

    sentry_sdk.flush()

    if span_streaming:
        assert len(items) == 1
        span = items[0].payload

        assert span["is_segment"] is True
        assert span["name"] == "generic WSGI request"
        assert span["attributes"]["sentry.op"] == "http.server"
        assert span["attributes"]["sentry.segment.name.source"] == "route"
        assert span["attributes"]["http.request.method"] == "GET"
        assert span["attributes"]["http.response.status_code"] == 200
        assert span["status"] == "ok"

        if send_pii:
            assert span["attributes"]["url.full"] == "http://localhost/dogs/are/great"
            assert span["attributes"]["url.path"] == "/dogs/are/great"
            assert span["attributes"]["http.query"] == "toy=tennisball"
        else:
            assert "url.path" not in span["attributes"]
            assert "url.full" not in span["attributes"]
            assert "http.query" not in span["attributes"]

    else:
        envelope = events[0]

        assert envelope["type"] == "transaction"
        assert envelope["transaction"] == "generic WSGI request"
        assert envelope["contexts"]["trace"]["op"] == "http.server"
        assert envelope["request"] == DictionaryContaining(
            {
                "method": "GET",
                "url": "http://localhost/dogs/are/great",
                "query_string": "toy=tennisball",
            }
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_has_trace_if_performance_enabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    def dogpark(environ, start_response):
        capture_message("Attempting to fetch the ball")
        raise ValueError("Fetch aborted. The ball was not returned.")

    sentry_init(
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(dogpark)
    client = Client(app)

    if span_streaming:
        items = capture_items("event", "span")
    else:
        events = capture_events()

    with pytest.raises(ValueError):
        client.get("http://dogs.are.great/sit/stay/rollover/")

    sentry_sdk.flush()

    if span_streaming:
        msg_event, error_event, span_item = items

        assert msg_event.type == "event"
        msg_event = msg_event.payload
        assert msg_event["contexts"]["trace"]
        assert "trace_id" in msg_event["contexts"]["trace"]

        assert error_event.type == "event"
        error_event = error_event.payload
        assert error_event["contexts"]["trace"]
        assert "trace_id" in error_event["contexts"]["trace"]

        assert span_item.type == "span"
        span_item = span_item.payload
        assert span_item["trace_id"] is not None

        assert (
            msg_event["contexts"]["trace"]["trace_id"]
            == error_event["contexts"]["trace"]["trace_id"]
            == span_item["trace_id"]
        )
    else:
        msg_event, error_event, transaction_event = events

        assert msg_event["contexts"]["trace"]
        assert "trace_id" in msg_event["contexts"]["trace"]

        assert error_event["contexts"]["trace"]
        assert "trace_id" in error_event["contexts"]["trace"]

        assert transaction_event["contexts"]["trace"]
        assert "trace_id" in transaction_event["contexts"]["trace"]

        assert (
            msg_event["contexts"]["trace"]["trace_id"]
            == error_event["contexts"]["trace"]["trace_id"]
            == transaction_event["contexts"]["trace"]["trace_id"]
        )


def test_has_trace_if_performance_disabled(
    sentry_init,
    capture_events,
):
    def dogpark(environ, start_response):
        capture_message("Attempting to fetch the ball")
        raise ValueError("Fetch aborted. The ball was not returned.")

    sentry_init()
    app = SentryWsgiMiddleware(dogpark)
    client = Client(app)
    events = capture_events()

    with pytest.raises(ValueError):
        client.get("http://dogs.are.great/sit/stay/rollover/")

    msg_event, error_event = events

    assert msg_event["contexts"]["trace"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert error_event["contexts"]["trace"]
    assert "trace_id" in error_event["contexts"]["trace"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_trace_from_headers_if_performance_enabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    def dogpark(environ, start_response):
        capture_message("Attempting to fetch the ball")
        raise ValueError("Fetch aborted. The ball was not returned.")

    sentry_init(
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(dogpark)
    client = Client(app)

    if span_streaming:
        items = capture_items("event", "span")
    else:
        events = capture_events()

    trace_id = "582b43a4192642f0b136d5159a501701"
    sentry_trace_header = "{}-{}-{}".format(trace_id, "6e8f22c393e68f19", 1)

    with pytest.raises(ValueError):
        client.get(
            "http://dogs.are.great/sit/stay/rollover/",
            headers={"sentry-trace": sentry_trace_header},
        )

    sentry_sdk.flush()

    if span_streaming:
        msg_event, error_event, span_item = items

        assert msg_event.payload["contexts"]["trace"]["trace_id"] == trace_id
        assert error_event.payload["contexts"]["trace"]["trace_id"] == trace_id
        assert span_item.payload["trace_id"] == trace_id
    else:
        msg_event, error_event, transaction_event = events

        assert msg_event["contexts"]["trace"]
        assert "trace_id" in msg_event["contexts"]["trace"]

        assert error_event["contexts"]["trace"]
        assert "trace_id" in error_event["contexts"]["trace"]

        assert transaction_event["contexts"]["trace"]
        assert "trace_id" in transaction_event["contexts"]["trace"]

        assert msg_event["contexts"]["trace"]["trace_id"] == trace_id
        assert error_event["contexts"]["trace"]["trace_id"] == trace_id
        assert transaction_event["contexts"]["trace"]["trace_id"] == trace_id


def test_trace_from_headers_if_performance_disabled(
    sentry_init,
    capture_events,
):
    def dogpark(environ, start_response):
        capture_message("Attempting to fetch the ball")
        raise ValueError("Fetch aborted. The ball was not returned.")

    sentry_init()
    app = SentryWsgiMiddleware(dogpark)
    client = Client(app)
    events = capture_events()

    trace_id = "582b43a4192642f0b136d5159a501701"
    sentry_trace_header = "{}-{}-{}".format(trace_id, "6e8f22c393e68f19", 1)

    with pytest.raises(ValueError):
        client.get(
            "http://dogs.are.great/sit/stay/rollover/",
            headers={"sentry-trace": sentry_trace_header},
        )

    msg_event, error_event = events

    assert msg_event["contexts"]["trace"]
    assert "trace_id" in msg_event["contexts"]["trace"]
    assert msg_event["contexts"]["trace"]["trace_id"] == trace_id

    assert error_event["contexts"]["trace"]
    assert "trace_id" in error_event["contexts"]["trace"]
    assert error_event["contexts"]["trace"]["trace_id"] == trace_id


@pytest.mark.parametrize("span_streaming", [True, False])
def test_traces_sampler_gets_correct_values_in_sampling_context(
    sentry_init,
    DictionaryContaining,  # noqa:N803
    span_streaming,
):
    def app(environ, start_response):
        start_response("200 OK", [])
        return ["Go get the ball! Good dog!"]

    traces_sampler = mock.Mock(return_value=True)
    sentry_init(
        send_default_pii=True,
        traces_sampler=traces_sampler,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(app)
    client = Client(app)

    client.get("/dogs/are/great/")

    traces_sampler.assert_any_call(
        DictionaryContaining(
            {
                "transaction_context": DictionaryContaining(
                    {
                        "name": "generic WSGI request",
                    },
                ),
                "wsgi_environ": DictionaryContaining(
                    {
                        "PATH_INFO": "/dogs/are/great/",
                        "REQUEST_METHOD": "GET",
                    },
                ),
            }
        )
    )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_session_mode_defaults_to_request_mode_in_wsgi_handler(
    capture_envelopes, sentry_init, span_streaming
):
    """
    Test that ensures that even though the default `session_mode` for
    auto_session_tracking is `application`, that flips to `request` when we are
    in the WSGI handler
    """

    def app(environ, start_response):
        start_response("200 OK", [])
        return ["Go get the ball! Good dog!"]

    traces_sampler = mock.Mock(return_value=True)
    sentry_init(
        send_default_pii=True,
        traces_sampler=traces_sampler,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(app)
    envelopes = capture_envelopes()

    client = Client(app)

    client.get("/dogs/are/great/")

    sentry_sdk.flush()

    session_envelopes = [
        e for e in envelopes if any(item.type == "sessions" for item in e.items)
    ]
    assert len(session_envelopes) == 1
    sess = session_envelopes[0]
    assert len(sess.items) == 1
    sess_event = sess.items[0].payload.json

    aggregates = sess_event["aggregates"]
    assert len(aggregates) == 1
    assert aggregates[0]["exited"] == 1


@pytest.mark.parametrize("span_streaming", [True, False])
def test_auto_session_tracking_with_aggregates(
    sentry_init, capture_envelopes, span_streaming
):
    """
    Test for correct session aggregates in auto session tracking.
    """

    def sample_app(environ, start_response):
        if environ["REQUEST_URI"] != "/dogs/are/great/":
            1 / 0

        start_response("200 OK", [])
        return ["Go get the ball! Good dog!"]

    traces_sampler = mock.Mock(return_value=True)
    sentry_init(
        send_default_pii=True,
        traces_sampler=traces_sampler,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(sample_app)
    envelopes = capture_envelopes()
    assert len(envelopes) == 0

    client = Client(app)
    client.get("/dogs/are/great/")
    client.get("/dogs/are/great/")
    try:
        client.get("/trigger/an/error/")
    except ZeroDivisionError:
        pass

    sentry_sdk.flush()

    count_item_types = Counter()
    for envelope in envelopes:
        for item in envelope.items:
            count_item_types[item.type] += 1

    if span_streaming:
        assert count_item_types["span"] == 3
    else:
        assert count_item_types["transaction"] == 3
    assert count_item_types["event"] == 1
    assert count_item_types["sessions"] == 1

    session_envelopes = [
        e for e in envelopes if any(item.type == "sessions" for item in e.items)
    ]
    assert len(session_envelopes) == 1
    session_aggregates = session_envelopes[0].items[0].payload.json["aggregates"]
    # Sum across buckets: sessions split by minute-truncated start time.
    assert sum(agg.get("exited", 0) for agg in session_aggregates) == 2
    assert sum(agg.get("crashed", 0) for agg in session_aggregates) == 1


@mock.patch("sentry_sdk.profiler.transaction_profiler.PROFILE_MINIMUM_SAMPLES", 0)
def test_profile_sent(
    sentry_init,
    capture_envelopes,
    teardown_profiling,
):
    def test_app(environ, start_response):
        start_response("200 OK", [])
        return ["Go get the ball! Good dog!"]

    sentry_init(
        traces_sample_rate=1.0,
        _experiments={"profiles_sample_rate": 1.0},
    )
    app = SentryWsgiMiddleware(test_app)
    envelopes = capture_envelopes()

    client = Client(app)
    client.get("/")

    envelopes = [envelope for envelope in envelopes]
    assert len(envelopes) == 1

    profiles = [item for item in envelopes[0].items if item.type == "profile"]
    assert len(profiles) == 1


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin_manual(sentry_init, capture_events, capture_items, span_streaming):
    def dogpark(environ, start_response):
        start_response("200 OK", [])
        return ["Go get the ball! Good dog!"]

    sentry_init(
        send_default_pii=True,
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(dogpark)

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = Client(app)
    client.get("/dogs/are/great/")

    sentry_sdk.flush()

    if span_streaming:
        assert len(items) == 1
        assert items[0].payload["attributes"]["sentry.origin"] == "manual"
    else:
        (event,) = events
        assert event["contexts"]["trace"]["origin"] == "manual"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin_custom(sentry_init, capture_events, capture_items, span_streaming):
    def dogpark(environ, start_response):
        start_response("200 OK", [])
        return ["Go get the ball! Good dog!"]

    sentry_init(
        send_default_pii=True,
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    app = SentryWsgiMiddleware(
        dogpark,
        span_origin="auto.dogpark.deluxe",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = Client(app)
    client.get("/dogs/are/great/")

    sentry_sdk.flush()

    if span_streaming:
        assert len(items) == 1
        assert items[0].payload["attributes"]["sentry.origin"] == "auto.dogpark.deluxe"
    else:
        (event,) = events
        assert event["contexts"]["trace"]["origin"] == "auto.dogpark.deluxe"


@pytest.mark.parametrize(
    "has_file_wrapper, has_fileno, expect_wrapped",
    [
        (True, True, False),  # both conditions met → unwrapped
        (False, True, True),  # no file_wrapper → wrapped
        (True, False, True),  # no fileno → wrapped
        (False, False, True),  # neither condition → wrapped
    ],
)
def test_file_response_wrapping(
    sentry_init, has_file_wrapper, has_fileno, expect_wrapped
):
    sentry_init()

    response_mock = mock.MagicMock()
    if not has_fileno:
        del response_mock.fileno

    def app(environ, start_response):
        start_response("200 OK", [])
        return response_mock

    environ_extra = {}
    if has_file_wrapper:
        environ_extra["wsgi.file_wrapper"] = mock.MagicMock()

    middleware = SentryWsgiMiddleware(app)

    result = middleware(
        {
            "REQUEST_METHOD": "GET",
            "PATH_INFO": "/",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
            "wsgi.url_scheme": "http",
            "wsgi.input": mock.MagicMock(),
            **environ_extra,
        },
        lambda status, headers: None,
    )

    if expect_wrapped:
        assert isinstance(result, _ScopedResponse)
    else:
        assert result is response_mock


@pytest.mark.parametrize(
    "environ,use_x_forwarded_for,expected_url",
    [
        # Without use_x_forwarded_for, wsgi.url_scheme is used
        (
            {
                "wsgi.url_scheme": "http",
                "SERVER_NAME": "example.com",
                "SERVER_PORT": "80",
                "PATH_INFO": "/test",
                "HTTP_X_FORWARDED_PROTO": "https",
            },
            False,
            "http://example.com/test",
        ),
        # With use_x_forwarded_for, HTTP_X_FORWARDED_PROTO is respected
        (
            {
                "wsgi.url_scheme": "http",
                "SERVER_NAME": "example.com",
                "SERVER_PORT": "80",
                "PATH_INFO": "/test",
                "HTTP_X_FORWARDED_PROTO": "https",
            },
            True,
            "https://example.com/test",
        ),
        # With use_x_forwarded_for but no forwarded proto, wsgi.url_scheme is used
        (
            {
                "wsgi.url_scheme": "http",
                "SERVER_NAME": "example.com",
                "SERVER_PORT": "80",
                "PATH_INFO": "/test",
            },
            True,
            "http://example.com/test",
        ),
        # Forwarded host with default https port is stripped using forwarded proto
        (
            {
                "wsgi.url_scheme": "http",
                "SERVER_NAME": "internal",
                "SERVER_PORT": "80",
                "PATH_INFO": "/test",
                "HTTP_X_FORWARDED_PROTO": "https",
                "HTTP_X_FORWARDED_HOST": "example.com:443",
            },
            True,
            "https://example.com/test",
        ),
        # Forwarded host with non-default port is preserved
        (
            {
                "wsgi.url_scheme": "http",
                "SERVER_NAME": "internal",
                "SERVER_PORT": "80",
                "PATH_INFO": "/test",
                "HTTP_X_FORWARDED_PROTO": "https",
                "HTTP_X_FORWARDED_HOST": "example.com:8443",
            },
            True,
            "https://example.com:8443/test",
        ),
        # Forwarded proto with HTTP_HOST (no forwarded host) strips default port
        (
            {
                "wsgi.url_scheme": "http",
                "HTTP_HOST": "example.com:443",
                "SERVER_NAME": "internal",
                "SERVER_PORT": "80",
                "PATH_INFO": "/test",
                "HTTP_X_FORWARDED_PROTO": "https",
            },
            True,
            "https://example.com/test",
        ),
    ],
    ids=[
        "ignores_forwarded_proto_when_disabled",
        "respects_forwarded_proto_when_enabled",
        "falls_back_to_url_scheme_when_no_forwarded_proto",
        "strips_default_https_port_from_forwarded_host",
        "preserves_non_default_port_on_forwarded_host",
        "strips_default_port_from_http_host_with_forwarded_proto",
    ],
)
def test_get_request_url_x_forwarded_proto(environ, use_x_forwarded_for, expected_url):
    assert get_request_url(environ, use_x_forwarded_for) == expected_url


@pytest.mark.parametrize("send_default_pii", [True, False])
def test_user_ip_address_on_all_spans(sentry_init, capture_items, send_default_pii):
    def dogpark(environ, start_response):
        with sentry_sdk.traces.start_span(name="child-span"):
            pass
        start_response("200 OK", [])
        return ["Go get the ball! Good dog!"]

    sentry_init(
        send_default_pii=send_default_pii,
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    app = SentryWsgiMiddleware(dogpark)
    client = Client(app)

    items = capture_items("span")

    client.get("/dogs/are/great/", environ_base={"REMOTE_ADDR": "127.0.0.1"})

    sentry_sdk.flush()

    child_span, server_span = [item.payload for item in items]

    if send_default_pii:
        assert server_span["attributes"]["user.ip_address"] == "127.0.0.1"
        assert child_span["attributes"]["user.ip_address"] == "127.0.0.1"
    else:
        assert "user.ip_address" not in server_span["attributes"]
        assert "user.ip_address" not in child_span["attributes"]



import sentry_sdk


def test_top_level_api(sentry_init, capture_items):
    sentry_init()

    items = capture_items("trace_metric")

    sentry_sdk.set_attribute("set", "value")
    sentry_sdk.set_attribute("removed", "value")
    sentry_sdk.remove_attribute("removed")
    # Attempting to remove a nonexistent attribute should not raise
    sentry_sdk.remove_attribute("nonexistent")

    sentry_sdk.metrics.count("test", 1)
    sentry_sdk.get_client().flush()

    metrics = [item.payload for item in items]
    (metric,) = metrics

    assert metric["attributes"]["set"] == "value"
    assert "removed" not in metric["attributes"]


def test_scope_precedence(sentry_init, capture_items):
    # Order of precedence, from most important to least:
    # 1. telemetry attributes (directly supplying attributes on creation or using set_attribute)
    # 2. current scope attributes
    # 3. isolation scope attributes
    # 4. global scope attributes
    sentry_init()

    items = capture_items("trace_metric")

    global_scope = sentry_sdk.get_global_scope()
    global_scope.set_attribute("global.attribute", "global")
    global_scope.set_attribute("overwritten.attribute", "global")

    isolation_scope = sentry_sdk.get_isolation_scope()
    isolation_scope.set_attribute("isolation.attribute", "isolation")
    isolation_scope.set_attribute("overwritten.attribute", "isolation")

    current_scope = sentry_sdk.get_current_scope()
    current_scope.set_attribute("current.attribute", "current")
    current_scope.set_attribute("overwritten.attribute", "current")

    sentry_sdk.metrics.count("test", 1)
    sentry_sdk.get_client().flush()

    metrics = [item.payload for item in items]
    (metric,) = metrics

    assert metric["attributes"]["global.attribute"] == "global"
    assert metric["attributes"]["isolation.attribute"] == "isolation"
    assert metric["attributes"]["current.attribute"] == "current"

    assert metric["attributes"]["overwritten.attribute"] == "current"


def test_telemetry_precedence(sentry_init, capture_items):
    # Order of precedence, from most important to least:
    # 1. telemetry attributes (directly supplying attributes on creation or using set_attribute)
    # 2. current scope attributes
    # 3. isolation scope attributes
    # 4. global scope attributes
    sentry_init()

    items = capture_items("trace_metric")

    global_scope = sentry_sdk.get_global_scope()
    global_scope.set_attribute("global.attribute", "global")
    global_scope.set_attribute("overwritten.attribute", "global")

    isolation_scope = sentry_sdk.get_isolation_scope()
    isolation_scope.set_attribute("isolation.attribute", "isolation")
    isolation_scope.set_attribute("overwritten.attribute", "isolation")

    current_scope = sentry_sdk.get_current_scope()
    current_scope.set_attribute("current.attribute", "current")
    current_scope.set_attribute("overwritten.attribute", "current")

    sentry_sdk.metrics.count(
        "test",
        1,
        attributes={
            "telemetry.attribute": "telemetry",
            "overwritten.attribute": "telemetry",
        },
    )

    sentry_sdk.get_client().flush()

    metrics = [item.payload for item in items]
    (metric,) = metrics

    assert metric["attributes"]["global.attribute"] == "global"
    assert metric["attributes"]["isolation.attribute"] == "isolation"
    assert metric["attributes"]["current.attribute"] == "current"
    assert metric["attributes"]["telemetry.attribute"] == "telemetry"

    assert metric["attributes"]["overwritten.attribute"] == "telemetry"


def test_attribute_out_of_scope(sentry_init, capture_items):
    sentry_init()

    items = capture_items("trace_metric")

    with sentry_sdk.new_scope() as scope:
        scope.set_attribute("outofscope.attribute", "out of scope")

    sentry_sdk.metrics.count("test", 1)

    sentry_sdk.get_client().flush()

    metrics = [item.payload for item in items]
    (metric,) = metrics

    assert "outofscope.attribute" not in metric["attributes"]


def test_remove_attribute(sentry_init, capture_items):
    sentry_init()

    items = capture_items("trace_metric")

    with sentry_sdk.new_scope() as scope:
        scope.set_attribute("some.attribute", 123)
        scope.remove_attribute("some.attribute")

        sentry_sdk.metrics.count("test", 1)

    sentry_sdk.get_client().flush()

    metrics = [item.payload for item in items]
    (metric,) = metrics

    assert "some.attribute" not in metric["attributes"]


def test_scope_attributes_preserialized(sentry_init, capture_items):
    def before_send_metric(metric, _):
        # Scope attrs show up serialized in before_send
        assert isinstance(metric["attributes"]["instance"], str)
        assert isinstance(metric["attributes"]["dictionary"], str)

        return metric

    sentry_init(before_send_metric=before_send_metric)

    items = capture_items("trace_metric")

    class Cat:
        pass

    instance = Cat()
    dictionary = {"color": "tortoiseshell"}

    with sentry_sdk.new_scope() as scope:
        scope.set_attribute("instance", instance)
        scope.set_attribute("dictionary", dictionary)

        # Scope attrs are stored preserialized
        assert isinstance(scope._attributes["instance"], str)
        assert isinstance(scope._attributes["dictionary"], str)

        sentry_sdk.metrics.count("test", 1)

    sentry_sdk.get_client().flush()

    metrics = [item.payload for item in items]
    (metric,) = metrics

    # Attrs originally from the scope are serialized when sent
    assert isinstance(metric["attributes"]["instance"], str)
    assert isinstance(metric["attributes"]["dictionary"], str)


def test_user_attributes(sentry_init, capture_items):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        _experiments={
            "trace_lifecycle": "stream",
        },
    )

    items = capture_items("trace_metric", "span")

    sentry_sdk.set_user(
        {
            "id": "123456",
            "username": "username",
            "email": "username@example.com",
            "ip_address": "127.0.0.1",
        }
    )

    with sentry_sdk.traces.start_span(name="login"):
        sentry_sdk.metrics.count("test", 1)

    sentry_sdk.get_client().flush()

    metric, span = [item.payload for item in items]

    assert metric["attributes"]["user.id"] == "123456"
    assert metric["attributes"]["user.name"] == "username"
    assert metric["attributes"]["user.email"] == "username@example.com"
    assert metric["attributes"]["user.ip_address"] == "127.0.0.1"

    assert span["attributes"]["user.id"] == "123456"
    assert span["attributes"]["user.name"] == "username"
    assert span["attributes"]["user.email"] == "username@example.com"
    assert span["attributes"]["user.ip_address"] == "127.0.0.1"



import asyncio
import logging
import os
import pickle
import socket
import sys
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest

try:
    import httpcore
except (ImportError, ModuleNotFoundError):
    httpcore = None

try:
    import gevent  # noqa: F401

    running_under_gevent = True
except ImportError:
    running_under_gevent = False

skip_under_gevent = pytest.mark.skipif(
    running_under_gevent, reason="Async tests not compatible with gevent"
)

import sentry_sdk
from sentry_sdk import (
    Client,
    Hub,
    add_breadcrumb,
    capture_message,
    get_isolation_scope,
    isolation_scope,
)
from sentry_sdk._compat import PY37, PY38
from sentry_sdk.envelope import Envelope, Item, PayloadRef, parse_json
from sentry_sdk.integrations.asyncio import AsyncioIntegration
from sentry_sdk.integrations.logging import LoggingIntegration, ignore_logger
from sentry_sdk.transport import (
    KEEP_ALIVE_SOCKET_OPTIONS,
    AsyncHttpTransport,
    HttpTransport,
    _parse_rate_limits,
)

server = None


def _make_async_transport_options(**overrides):
    defaults = {
        "dsn": "https://foo@sentry.io/123",
        "transport": None,
        "_experiments": {"transport_async": True},
        "integrations": [AsyncioIntegration()],
        "send_client_reports": True,
        "transport_queue_size": 100,
        "keep_alive": False,
        "socket_options": None,
        "ca_certs": None,
        "cert_file": None,
        "key_file": None,
        "http_proxy": None,
        "https_proxy": None,
        "proxy_headers": None,
    }
    defaults.update(overrides)
    return defaults


@pytest.fixture(scope="module", autouse=True)
def make_capturing_server(request, wsgi_capturing_server):
    global server
    server = wsgi_capturing_server
    server.start()
    request.addfinalizer(server.stop)


@pytest.fixture
def capturing_server():
    global server
    server.clear_captured()
    return server


@pytest.fixture
def make_client(request, capturing_server):
    def inner(**kwargs):
        return Client(
            "http://foobar@{}/132".format(capturing_server.url[len("http://") :]),
            **kwargs,
        )

    return inner


def mock_transaction_envelope(span_count: int) -> "Envelope":
    event = defaultdict(
        mock.MagicMock,
        type="transaction",
        spans=[mock.MagicMock() for _ in range(span_count)],
    )

    envelope = Envelope()
    envelope.add_transaction(event)

    return envelope


@pytest.mark.parametrize("debug", (True, False))
@pytest.mark.parametrize("client_flush_method", ["close", "flush"])
@pytest.mark.parametrize("use_pickle", (True, False))
@pytest.mark.parametrize("compression_level", (0, 9, None))
@pytest.mark.parametrize(
    "compression_algo",
    (("gzip", "br", "<invalid>", None) if PY37 else ("gzip", "<invalid>", None)),
)
@pytest.mark.parametrize("http2", [True, False] if PY38 else [False])
def test_transport_works(
    capturing_server,
    request,
    capsys,
    caplog,
    debug,
    make_client,
    client_flush_method,
    use_pickle,
    compression_level,
    compression_algo,
    http2,
):
    caplog.set_level(logging.DEBUG)

    experiments = {}
    if compression_level is not None:
        experiments["transport_compression_level"] = compression_level

    if compression_algo is not None:
        experiments["transport_compression_algo"] = compression_algo

    if http2:
        experiments["transport_http2"] = True

    client = make_client(
        debug=debug,
        _experiments=experiments,
    )

    if use_pickle:
        client = pickle.loads(pickle.dumps(client))

    sentry_sdk.get_global_scope().set_client(client)
    request.addfinalizer(lambda: sentry_sdk.get_global_scope().set_client(None))

    add_breadcrumb(
        level="info", message="i like bread", timestamp=datetime.now(timezone.utc)
    )
    capture_message("löl")

    getattr(client, client_flush_method)()

    out, err = capsys.readouterr()
    assert not err and not out
    assert capturing_server.captured
    should_compress = (
        # default is to compress with brotli if available, gzip otherwise
        (compression_level is None)
        or (
            # setting compression level to 0 means don't compress
            compression_level > 0
        )
    ) and (
        # if we couldn't resolve to a known algo, we don't compress
        compression_algo != "<invalid>"
    )

    assert capturing_server.captured[0].compressed == should_compress

    assert any("Sending envelope" in record.msg for record in caplog.records) == debug


@pytest.mark.parametrize(
    "num_pools,expected_num_pools",
    (
        (None, 2),
        (2, 2),
        (10, 10),
    ),
)
def test_transport_num_pools(make_client, num_pools, expected_num_pools):
    _experiments = {}
    if num_pools is not None:
        _experiments["transport_num_pools"] = num_pools

    client = make_client(_experiments=_experiments)

    options = client.transport._get_pool_options()
    assert options["num_pools"] == expected_num_pools


@pytest.mark.parametrize(
    "http2", [True, False] if sys.version_info >= (3, 8) else [False]
)
def test_two_way_ssl_authentication(make_client, http2):
    _experiments = {}
    if http2:
        _experiments["transport_http2"] = True

    current_dir = os.path.dirname(__file__)
    cert_file = f"{current_dir}/test.pem"
    key_file = f"{current_dir}/test.key"
    client = make_client(
        cert_file=cert_file,
        key_file=key_file,
        _experiments=_experiments,
    )
    options = client.transport._get_pool_options()

    if http2:
        assert options["ssl_context"] is not None
    else:
        assert options["cert_file"] == cert_file
        assert options["key_file"] == key_file


def test_socket_options(make_client):
    socket_options = [
        (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
        (socket.SOL_TCP, socket.TCP_KEEPINTVL, 10),
        (socket.SOL_TCP, socket.TCP_KEEPCNT, 6),
    ]

    client = make_client(socket_options=socket_options)

    options = client.transport._get_pool_options()
    assert options["socket_options"] == socket_options


def test_keep_alive_true(make_client):
    client = make_client(keep_alive=True)

    options = client.transport._get_pool_options()
    assert options["socket_options"] == KEEP_ALIVE_SOCKET_OPTIONS


def test_keep_alive_on_by_default(make_client):
    client = make_client()
    options = client.transport._get_pool_options()
    assert "socket_options" not in options


def test_default_timeout(make_client):
    client = make_client()

    options = client.transport._get_pool_options()
    assert "timeout" in options
    assert options["timeout"].total == client.transport.TIMEOUT


@pytest.mark.skipif(not PY38, reason="HTTP2 libraries are only available in py3.8+")
def test_default_timeout_http2(make_client):
    client = make_client(_experiments={"transport_http2": True})

    with mock.patch(
        "sentry_sdk.transport.httpcore.ConnectionPool.request",
        return_value=httpcore.Response(200),
    ) as request_mock:
        sentry_sdk.get_global_scope().set_client(client)
        capture_message("hi")
        client.flush()

    request_mock.assert_called_once()
    assert request_mock.call_args.kwargs["extensions"] == {
        "timeout": {
            "pool": client.transport.TIMEOUT,
            "connect": client.transport.TIMEOUT,
            "write": client.transport.TIMEOUT,
            "read": client.transport.TIMEOUT,
        }
    }


@pytest.mark.skipif(not PY38, reason="HTTP2 libraries are only available in py3.8+")
def test_http2_with_https_dsn(make_client):
    client = make_client(_experiments={"transport_http2": True})
    client.transport.parsed_dsn.scheme = "https"
    options = client.transport._get_pool_options()
    assert options["http2"] is True


@pytest.mark.skipif(not PY38, reason="HTTP2 libraries are only available in py3.8+")
def test_no_http2_with_http_dsn(make_client):
    client = make_client(_experiments={"transport_http2": True})
    client.transport.parsed_dsn.scheme = "http"
    options = client.transport._get_pool_options()
    assert options["http2"] is False


def test_socket_options_override_keep_alive(make_client):
    socket_options = [
        (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
        (socket.SOL_TCP, socket.TCP_KEEPINTVL, 10),
        (socket.SOL_TCP, socket.TCP_KEEPCNT, 6),
    ]

    client = make_client(socket_options=socket_options, keep_alive=False)

    options = client.transport._get_pool_options()
    assert options["socket_options"] == socket_options


def test_socket_options_merge_with_keep_alive(make_client):
    socket_options = [
        (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 42),
        (socket.SOL_TCP, socket.TCP_KEEPINTVL, 42),
    ]

    client = make_client(socket_options=socket_options, keep_alive=True)

    options = client.transport._get_pool_options()
    try:
        assert options["socket_options"] == [
            (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 42),
            (socket.SOL_TCP, socket.TCP_KEEPINTVL, 42),
            (socket.SOL_TCP, socket.TCP_KEEPIDLE, 45),
            (socket.SOL_TCP, socket.TCP_KEEPCNT, 6),
        ]
    except AttributeError:
        assert options["socket_options"] == [
            (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 42),
            (socket.SOL_TCP, socket.TCP_KEEPINTVL, 42),
            (socket.SOL_TCP, socket.TCP_KEEPCNT, 6),
        ]


def test_socket_options_override_defaults(make_client):
    # If socket_options are set to [], this doesn't mean the user doesn't want
    # any custom socket_options, but rather that they want to disable the urllib3
    # socket option defaults, so we need to set this and not ignore it.
    client = make_client(socket_options=[])

    options = client.transport._get_pool_options()
    assert options["socket_options"] == []


def test_transport_infinite_loop(capturing_server, request, make_client):
    client = make_client(
        debug=True,
        # Make sure we cannot create events from our own logging
        integrations=[LoggingIntegration(event_level=logging.DEBUG)],
    )

    # I am not sure why, but "werkzeug" logger makes an INFO log on sending
    # the message "hi" and does creates an infinite look.
    # Ignoring this for breaking the infinite loop and still we can test
    # that our own log messages (sent from `_IGNORED_LOGGERS`) are not leading
    # to an infinite loop
    ignore_logger("werkzeug")

    sentry_sdk.get_global_scope().set_client(client)
    with isolation_scope():
        capture_message("hi")
        client.flush()

    assert len(capturing_server.captured) == 1


def test_transport_no_thread_on_shutdown_no_errors(capturing_server, make_client):
    client = make_client()

    # make it seem like the interpreter is shutting down
    with mock.patch(
        "threading.Thread.start",
        side_effect=RuntimeError("can't create new thread at interpreter shutdown"),
    ):
        sentry_sdk.get_global_scope().set_client(client)
        with isolation_scope():
            capture_message("hi")

    # nothing exploded but also no events can be sent anymore
    assert len(capturing_server.captured) == 0


NOW = datetime(2014, 6, 2)


@pytest.mark.parametrize(
    "input,expected",
    [
        # Invalid rate limits
        ("", {}),
        ("invalid", {}),
        (",,,", {}),
        (
            "42::organization, invalid, 4711:foobar;transaction;security:project",
            {
                None: NOW + timedelta(seconds=42),
                "transaction": NOW + timedelta(seconds=4711),
                "security": NOW + timedelta(seconds=4711),
                # Unknown data categories
                "foobar": NOW + timedelta(seconds=4711),
            },
        ),
        (
            "4711:foobar;;transaction:organization",
            {
                "transaction": NOW + timedelta(seconds=4711),
                # Unknown data categories
                "foobar": NOW + timedelta(seconds=4711),
                "": NOW + timedelta(seconds=4711),
            },
        ),
    ],
)
def test_parse_rate_limits(input, expected):
    assert dict(_parse_rate_limits(input, now=NOW)) == expected


def test_envelope_too_large_response(capturing_server, make_client):
    client = make_client()

    capturing_server.respond_with(code=413)
    client.capture_event({"type": "error"})
    client.capture_event({"type": "transaction"})
    client.flush()

    # Error, transaction, and client report payloads
    assert len(capturing_server.captured) == 3
    report = parse_json(capturing_server.captured[2].envelope.items[0].get_bytes())

    # Client reports for error, transaction and included span
    assert len(report["discarded_events"]) == 3
    assert {"reason": "send_error", "category": "error", "quantity": 1} in report[
        "discarded_events"
    ]
    assert {"reason": "send_error", "category": "span", "quantity": 1} in report[
        "discarded_events"
    ]
    assert {"reason": "send_error", "category": "transaction", "quantity": 1} in report[
        "discarded_events"
    ]

    capturing_server.clear_captured()


def test_simple_rate_limits(capturing_server, make_client):
    client = make_client()
    capturing_server.respond_with(code=429, headers={"Retry-After": "4"})

    client.capture_event({"type": "transaction"})
    client.flush()

    assert len(capturing_server.captured) == 1
    assert capturing_server.captured[0].path == "/api/132/envelope/"
    capturing_server.clear_captured()

    assert set(client.transport._disabled_until) == set([None])

    client.capture_event({"type": "transaction"})
    client.capture_event({"type": "event"})
    client.flush()

    assert not capturing_server.captured


@pytest.mark.parametrize("response_code", [200, 429])
def test_data_category_limits(
    capturing_server, response_code, make_client, monkeypatch
):
    client = make_client(send_client_reports=False)

    captured_outcomes = []

    def record_lost_event(reason, data_category=None, item=None):
        if data_category is None:
            data_category = item.data_category
        return captured_outcomes.append((reason, data_category))

    monkeypatch.setattr(client.transport, "record_lost_event", record_lost_event)

    capturing_server.respond_with(
        code=response_code,
        headers={"X-Sentry-Rate-Limits": "4711:transaction:organization"},
    )

    client.capture_event({"type": "transaction"})
    client.flush()

    assert len(capturing_server.captured) == 1
    assert capturing_server.captured[0].path == "/api/132/envelope/"
    capturing_server.clear_captured()

    assert set(client.transport._disabled_until) == set(["transaction"])

    client.capture_event({"type": "transaction"})
    client.capture_event({"type": "transaction"})
    client.flush()

    assert not capturing_server.captured

    client.capture_event({"type": "event"})
    client.flush()

    assert len(capturing_server.captured) == 1
    assert capturing_server.captured[0].path == "/api/132/envelope/"

    assert captured_outcomes == [
        ("ratelimit_backoff", "transaction"),
        ("ratelimit_backoff", "transaction"),
    ]


@pytest.mark.parametrize("response_code", [200, 429])
def test_data_category_limits_reporting(
    capturing_server, response_code, make_client, monkeypatch
):
    client = make_client(send_client_reports=True)

    capturing_server.respond_with(
        code=response_code,
        headers={
            "X-Sentry-Rate-Limits": "4711:transaction:organization, 4711:attachment:organization"
        },
    )

    outcomes_enabled = False
    real_fetch = client.transport._fetch_pending_client_report

    def intercepting_fetch(*args, **kwargs):
        if outcomes_enabled:
            return real_fetch(*args, **kwargs)

    monkeypatch.setattr(
        client.transport, "_fetch_pending_client_report", intercepting_fetch
    )
    # get rid of threading making things hard to track
    monkeypatch.setattr(client.transport._worker, "submit", lambda x: x() or True)

    client.capture_event({"type": "transaction"})
    client.flush()

    assert len(capturing_server.captured) == 1
    assert capturing_server.captured[0].path == "/api/132/envelope/"
    capturing_server.clear_captured()

    assert set(client.transport._disabled_until) == set(["attachment", "transaction"])

    client.capture_event({"type": "transaction"})
    client.capture_event({"type": "transaction"})
    capturing_server.clear_captured()

    # flush out the events but don't flush the client reports
    client.flush()
    client.transport._last_client_report_sent = 0
    outcomes_enabled = True

    scope = get_isolation_scope()
    scope.add_attachment(bytes=b"Hello World", filename="hello.txt")
    client.capture_event({"type": "error"}, scope=scope)
    client.flush()

    # this goes out with an extra envelope because it's flushed after the last item
    # that is normally in the queue.  This is quite funny in a way because it means
    # that the envelope that caused its own over quota report (an error with an
    # attachment) will include its outcome since it's pending.
    assert len(capturing_server.captured) == 1
    envelope = capturing_server.captured[0].envelope
    assert envelope.items[0].type == "event"
    assert envelope.items[1].type == "client_report"
    report = parse_json(envelope.items[1].get_bytes())

    discarded_events = report["discarded_events"]

    assert len(discarded_events) == 3
    assert {
        "category": "transaction",
        "reason": "ratelimit_backoff",
        "quantity": 2,
    } in discarded_events
    assert {
        "category": "span",
        "reason": "ratelimit_backoff",
        "quantity": 2,
    } in discarded_events
    assert {
        "category": "attachment",
        "reason": "ratelimit_backoff",
        "quantity": 11,
    } in discarded_events

    capturing_server.clear_captured()

    # here we sent a normal event
    client.capture_event({"type": "transaction"})
    client.capture_event({"type": "error", "release": "foo"})
    client.flush()

    assert len(capturing_server.captured) == 2

    assert len(capturing_server.captured[0].envelope.items) == 1
    event = capturing_server.captured[0].envelope.items[0].get_event()
    assert event["type"] == "error"
    assert event["release"] == "foo"

    envelope = capturing_server.captured[1].envelope
    assert envelope.items[0].type == "client_report"
    report = parse_json(envelope.items[0].get_bytes())

    discarded_events = report["discarded_events"]
    assert len(discarded_events) == 2
    assert {
        "category": "transaction",
        "reason": "ratelimit_backoff",
        "quantity": 1,
    } in discarded_events
    assert {
        "category": "span",
        "reason": "ratelimit_backoff",
        "quantity": 1,
    } in discarded_events


@pytest.mark.parametrize("response_code", [200, 429])
def test_complex_limits_without_data_category(
    capturing_server, response_code, make_client
):
    client = make_client()
    capturing_server.respond_with(
        code=response_code,
        headers={"X-Sentry-Rate-Limits": "4711::organization"},
    )

    client.capture_event({"type": "transaction"})
    client.flush()

    assert len(capturing_server.captured) == 1
    assert capturing_server.captured[0].path == "/api/132/envelope/"
    capturing_server.clear_captured()

    assert set(client.transport._disabled_until) == set([None])

    client.capture_event({"type": "transaction"})
    client.capture_event({"type": "transaction"})
    client.capture_event({"type": "event"})
    client.flush()

    assert len(capturing_server.captured) == 0


@pytest.mark.parametrize("response_code", [200, 429])
@pytest.mark.parametrize(
    "item",
    [
        Item(payload=b"{}", type="log"),
        Item(
            type="log",
            content_type="application/vnd.sentry.items.log+json",
            headers={
                "item_count": 2,
            },
            payload=PayloadRef(
                json={
                    "items": [
                        {
                            "body": "This is a 'info' log...",
                            "level": "info",
                            "timestamp": datetime(
                                2025, 1, 1, tzinfo=timezone.utc
                            ).timestamp(),
                            "trace_id": "00000000-0000-0000-0000-000000000000",
                            "attributes": {
                                "sentry.environment": {
                                    "value": "production",
                                    "type": "string",
                                },
                                "sentry.release": {
                                    "value": "1.0.0",
                                    "type": "string",
                                },
                                "sentry.sdk.name": {
                                    "value": "sentry.python",
                                    "type": "string",
                                },
                                "sentry.sdk.version": {
                                    "value": "2.45.0",
                                    "type": "string",
                                },
                                "sentry.severity_number": {
                                    "value": 9,
                                    "type": "integer",
                                },
                                "sentry.severity_text": {
                                    "value": "info",
                                    "type": "string",
                                },
                                "server.address": {
                                    "value": "test-server",
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "body": "The recorded value was '2.0'",
                            "level": "warn",
                            "timestamp": datetime(
                                2025, 1, 1, tzinfo=timezone.utc
                            ).timestamp(),
                            "trace_id": "00000000-0000-0000-0000-000000000000",
                            "attributes": {
                                "sentry.message.parameter.float_var": {
                                    "value": 2.0,
                                    "type": "double",
                                },
                                "sentry.message.template": {
                                    "value": "The recorded value was '{float_var}'",
                                    "type": "string",
                                },
                                "sentry.sdk.name": {
                                    "value": "sentry.python",
                                    "type": "string",
                                },
                                "sentry.sdk.version": {
                                    "value": "2.45.0",
                                    "type": "string",
                                },
                                "server.address": {
                                    "value": "test-server",
                                    "type": "string",
                                },
                                "sentry.environment": {
                                    "value": "production",
                                    "type": "string",
                                },
                                "sentry.release": {
                                    "value": "1.0.0",
                                    "type": "string",
                                },
                                "sentry.severity_number": {
                                    "value": 13,
                                    "type": "integer",
                                },
                                "sentry.severity_text": {
                                    "value": "warn",
                                    "type": "string",
                                },
                            },
                        },
                    ]
                }
            ),
        ),
    ],
)
def test_log_item_limits(capturing_server, response_code, item, make_client):
    client = make_client()
    capturing_server.respond_with(
        code=response_code,
        headers={
            "X-Sentry-Rate-Limits": "4711:log_item:organization:quota_exceeded:custom"
        },
    )

    envelope = Envelope()
    envelope.add_item(item)
    client.transport.capture_envelope(envelope)
    client.flush()

    assert len(capturing_server.captured) == 1
    assert capturing_server.captured[0].path == "/api/132/envelope/"
    capturing_server.clear_captured()

    assert set(client.transport._disabled_until) == {"log_item"}

    client.transport.capture_envelope(envelope)
    client.capture_event({"type": "transaction"})
    client.flush()

    assert len(capturing_server.captured) == 2

    envelope = capturing_server.captured[0].envelope
    assert envelope.items[0].type == "transaction"
    envelope = capturing_server.captured[1].envelope
    assert envelope.items[0].type == "client_report"
    report = parse_json(envelope.items[0].get_bytes())

    assert {
        "category": "log_item",
        "reason": "ratelimit_backoff",
        "quantity": 1,
    } in report["discarded_events"]

    expected_lost_bytes = 1243
    if item.payload.bytes == b"{}":
        expected_lost_bytes = 2

    assert {
        "category": "log_byte",
        "reason": "ratelimit_backoff",
        "quantity": expected_lost_bytes,
    } in report["discarded_events"]


def test_hub_cls_backwards_compat():
    class TestCustomHubClass(Hub):
        pass

    transport = HttpTransport(
        defaultdict(lambda: None, {"dsn": "https://123abc@example.com/123"})
    )

    with pytest.deprecated_call():
        assert transport.hub_cls is Hub

    with pytest.deprecated_call():
        transport.hub_cls = TestCustomHubClass

    with pytest.deprecated_call():
        assert transport.hub_cls is TestCustomHubClass


@pytest.mark.parametrize("quantity", (1, 2, 10))
def test_record_lost_event_quantity(capturing_server, make_client, quantity):
    client = make_client()
    transport = client.transport

    transport.record_lost_event(reason="test", data_category="span", quantity=quantity)
    client.flush()

    (captured,) = capturing_server.captured  # Should only be one envelope
    envelope = captured.envelope
    (item,) = envelope.items  # Envelope should only have one item

    assert item.type == "client_report"

    report = parse_json(item.get_bytes())

    assert report["discarded_events"] == [
        {"category": "span", "reason": "test", "quantity": quantity}
    ]


@pytest.mark.parametrize("span_count", (0, 1, 2, 10))
def test_record_lost_event_transaction_item(capturing_server, make_client, span_count):
    client = make_client()
    transport = client.transport

    envelope = mock_transaction_envelope(span_count)
    (transaction_item,) = envelope.items

    transport.record_lost_event(reason="test", item=transaction_item)
    client.flush()

    (captured,) = capturing_server.captured  # Should only be one envelope
    envelope = captured.envelope
    (item,) = envelope.items  # Envelope should only have one item

    assert item.type == "client_report"

    report = parse_json(item.get_bytes())
    discarded_events = report["discarded_events"]

    assert len(discarded_events) == 2

    assert {
        "category": "transaction",
        "reason": "test",
        "quantity": 1,
    } in discarded_events

    assert {
        "category": "span",
        "reason": "test",
        "quantity": span_count + 1,
    } in discarded_events


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.parametrize("debug", (True, False))
@pytest.mark.parametrize("client_flush_method", ["close", "flush"])
@pytest.mark.parametrize("use_pickle", (True, False))
@pytest.mark.parametrize("compression_level", (0, 9, None))
@pytest.mark.parametrize("compression_algo", ("gzip", "br", "<invalid>", None))
@pytest.mark.skipif(not PY38, reason="Async transport only supported in Python 3.8+")
async def test_transport_works_async(
    capturing_server,
    request,
    capsys,
    caplog,
    debug,
    make_client,
    client_flush_method,
    use_pickle,
    compression_level,
    compression_algo,
):
    caplog.set_level(logging.DEBUG)

    experiments = {}
    if compression_level is not None:
        experiments["transport_compression_level"] = compression_level

    if compression_algo is not None:
        experiments["transport_compression_algo"] = compression_algo

    # Enable async transport
    experiments["transport_async"] = True

    client = make_client(
        debug=debug,
        _experiments=experiments,
        integrations=[AsyncioIntegration()],
    )

    if use_pickle:
        client = pickle.loads(pickle.dumps(client))

    # Verify we're using async transport
    assert isinstance(client.transport, AsyncHttpTransport), (
        "Expected AsyncHttpTransport"
    )

    sentry_sdk.get_global_scope().set_client(client)
    request.addfinalizer(lambda: sentry_sdk.get_global_scope().set_client(None))

    add_breadcrumb(
        level="info", message="i like bread", timestamp=datetime.now(timezone.utc)
    )
    capture_message("löl")

    if client_flush_method == "close":
        await client.close_async(timeout=2.0)
    if client_flush_method == "flush":
        await client.flush_async(timeout=2.0)

    out, err = capsys.readouterr()
    assert not err and not out
    assert capturing_server.captured
    should_compress = (
        # default is to compress with brotli if available, gzip otherwise
        (compression_level is None)
        or (
            # setting compression level to 0 means don't compress
            compression_level > 0
        )
    ) and (
        # if we couldn't resolve to a known algo, we don't compress
        compression_algo != "<invalid>"
    )

    assert capturing_server.captured[0].compressed == should_compress
    # After flush, the worker task is still running, but the end of the test will shut down the event loop
    # Therefore, we need to explicitly close the client to clean up the worker task
    assert any("Sending envelope" in record.msg for record in caplog.records) == debug
    if client_flush_method == "flush":
        await client.close_async(timeout=2.0)


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async transport requires Python 3.8+")
async def test_async_two_way_ssl_authentication():
    current_dir = os.path.dirname(__file__)
    cert_file = f"{current_dir}/test.pem"
    key_file = f"{current_dir}/test.key"

    client = Client(
        "https://foo@sentry.io/123",
        cert_file=cert_file,
        key_file=key_file,
        _experiments={"transport_async": True},
        integrations=[AsyncioIntegration()],
    )
    assert isinstance(client.transport, AsyncHttpTransport)

    options = client.transport._get_pool_options()
    assert options["ssl_context"] is not None

    await client.close_async()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async transport requires Python 3.8+")
async def test_async_transport_concurrent_requests(
    capturing_server, make_client, caplog
):
    """Test multiple simultaneous envelope submissions"""
    caplog.set_level(logging.DEBUG)
    client = make_client(
        _experiments={"transport_async": True}, integrations=[AsyncioIntegration()]
    )
    assert isinstance(client.transport, AsyncHttpTransport)
    sentry_sdk.get_global_scope().set_client(client)

    num_messages = 15

    async def send_message(i):
        capture_message(f"concurrent message {i}")

    tasks = [send_message(i) for i in range(num_messages)]
    await asyncio.gather(*tasks)
    await client.close_async(timeout=2.0)
    assert len(capturing_server.captured) == num_messages


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async transport requires Python 3.8+")
async def test_async_transport_rate_limiting_with_concurrency(
    capturing_server, make_client, request
):
    """Test async transport rate limiting with concurrent requests"""
    client = make_client(
        _experiments={"transport_async": True}, integrations=[AsyncioIntegration()]
    )

    assert isinstance(client.transport, AsyncHttpTransport)
    sentry_sdk.get_global_scope().set_client(client)
    request.addfinalizer(lambda: sentry_sdk.get_global_scope().set_client(None))
    capturing_server.respond_with(
        code=429, headers={"X-Sentry-Rate-Limits": "60:error:organization"}
    )

    # Send one request first to trigger rate limiting
    capture_message("initial message")
    await asyncio.sleep(0.1)  # Wait for request to execute
    assert client.transport._check_disabled("error") is True
    capturing_server.clear_captured()

    async def send_message(i):
        capture_message(f"message {i}")
        await asyncio.sleep(0.01)

    await asyncio.gather(*[send_message(i) for i in range(5)])
    await asyncio.sleep(0.1)
    # New request should be dropped due to rate limiting
    assert len(capturing_server.captured) == 0
    await client.close_async(timeout=2.0)



"""
Tests need pytest-asyncio installed.

Tests need a local postgresql instance running, this can best be done using
```sh
docker run --rm --name some-postgres -e POSTGRES_USER=postgres -e POSTGRES_PASSWORD=sentry -d -p 5432:5432 postgres
```

The tests use the following credentials to establish a database connection.
"""

import datetime
import os
from contextlib import contextmanager
from unittest import mock

import asyncpg
import pytest
import pytest_asyncio
from asyncpg import Connection, connect

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations.asyncpg import AsyncPGIntegration
from sentry_sdk.tracing_utils import record_sql_queries
from tests.conftest import ApproxDict

PG_HOST = os.getenv("SENTRY_PYTHON_TEST_POSTGRES_HOST", "localhost")
PG_PORT = int(os.getenv("SENTRY_PYTHON_TEST_POSTGRES_PORT", "5432"))
PG_USER = os.getenv("SENTRY_PYTHON_TEST_POSTGRES_USER", "postgres")
PG_PASSWORD = os.getenv("SENTRY_PYTHON_TEST_POSTGRES_PASSWORD", "sentry")
PG_NAME_BASE = os.getenv("SENTRY_PYTHON_TEST_POSTGRES_NAME", "postgres")


def _get_db_name():
    pid = os.getpid()
    return f"{PG_NAME_BASE}_{pid}"


PG_NAME = _get_db_name()

PG_CONNECTION_URI = "postgresql://{}:{}@{}/{}".format(
    PG_USER, PG_PASSWORD, PG_HOST, PG_NAME
)
CRUMBS_CONNECT = {
    "category": "query",
    "data": ApproxDict(
        {
            "db.name": PG_NAME,
            "db.system": "postgresql",
            "db.user": PG_USER,
            "db.driver.name": "asyncpg",
            "server.address": PG_HOST,
            "server.port": PG_PORT,
        }
    ),
    "message": "connect",
    "type": "default",
}
CRUMBS_CONNECT_STREAMING = {
    "category": "query",
    "data": ApproxDict(
        {
            "sentry.op": "db",
            "sentry.origin": "auto.db.asyncpg",
            "db.system.name": "postgresql",
            "db.namespace": PG_NAME,
            "db.user": PG_USER,
            "db.driver.name": "asyncpg",
            "server.address": PG_HOST,
            "server.port": PG_PORT,
        }
    ),
    "message": "connect",
    "type": "default",
}


@pytest_asyncio.fixture(autouse=True)
async def _clean_pg():
    # Create the test database if it doesn't exist
    default_conn = await connect(
        "postgresql://{}:{}@{}".format(PG_USER, PG_PASSWORD, PG_HOST)
    )
    try:
        # Check if database exists, create if not
        result = await default_conn.fetchval(
            "SELECT 1 FROM pg_database WHERE datname = $1", PG_NAME
        )
        if not result:
            await default_conn.execute(f'CREATE DATABASE "{PG_NAME}"')
    finally:
        await default_conn.close()

    # Now connect to our test database and set up the table
    conn = await connect(PG_CONNECTION_URI)
    await conn.execute("DROP TABLE IF EXISTS users")
    await conn.execute(
        """
            CREATE TABLE users(
                id serial PRIMARY KEY,
                name text,
                password text,
                dob date
            )
        """
    )
    await conn.close()


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_connect(
    sentry_init, capture_events, capture_items, span_streaming
) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration()],
        _experiments={
            "record_sql_params": True,
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("event")
    else:
        events = capture_events()

    conn: Connection = await connect(PG_CONNECTION_URI)

    await conn.close()

    capture_message("hi")

    if span_streaming:
        event = items[0].payload
    else:
        (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    expected_crumbs_connect = (
        CRUMBS_CONNECT_STREAMING if span_streaming else CRUMBS_CONNECT
    )
    assert event["breadcrumbs"]["values"] == [expected_crumbs_connect]


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_execute(
    sentry_init, capture_events, capture_items, span_streaming
) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration()],
        _experiments={
            "record_sql_params": True,
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("event")
    else:
        events = capture_events()

    conn: Connection = await connect(PG_CONNECTION_URI)

    await conn.execute(
        "INSERT INTO users(name, password, dob) VALUES ('Alice', 'pw', '1990-12-25')",
    )

    await conn.execute(
        "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
        "Bob",
        "secret_pw",
        datetime.date(1984, 3, 1),
    )

    row = await conn.fetchrow("SELECT * FROM users WHERE name = $1", "Bob")
    assert row == (2, "Bob", "secret_pw", datetime.date(1984, 3, 1))

    row = await conn.fetchrow("SELECT * FROM users WHERE name = 'Bob'")
    assert row == (2, "Bob", "secret_pw", datetime.date(1984, 3, 1))

    await conn.close()

    capture_message("hi")

    if span_streaming:
        event = items[0].payload
    else:
        (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    expected_crumbs_connect = (
        CRUMBS_CONNECT_STREAMING if span_streaming else CRUMBS_CONNECT
    )
    assert event["breadcrumbs"]["values"] == [
        expected_crumbs_connect,
        {
            "category": "query",
            "data": {},
            "message": "INSERT INTO users(name, password, dob) VALUES ('Alice', 'pw', '1990-12-25')",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "SELECT * FROM users WHERE name = $1",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "SELECT * FROM users WHERE name = 'Bob'",
            "type": "default",
        },
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_execute_many(
    sentry_init, capture_events, capture_items, span_streaming
) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration()],
        _experiments={
            "record_sql_params": True,
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("event")
    else:
        events = capture_events()

    conn: Connection = await connect(PG_CONNECTION_URI)

    await conn.executemany(
        "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
        [
            ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
            ("Alice", "pw", datetime.date(1990, 12, 25)),
        ],
    )

    await conn.close()

    capture_message("hi")

    if span_streaming:
        event = items[0].payload
    else:
        (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    expected_crumbs_connect = (
        CRUMBS_CONNECT_STREAMING if span_streaming else CRUMBS_CONNECT
    )
    assert event["breadcrumbs"]["values"] == [
        expected_crumbs_connect,
        {
            "category": "query",
            "data": {"db.executemany": True},
            "message": "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
            "type": "default",
        },
    ]


@pytest.mark.asyncio
async def test_record_params(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration(record_params=True)],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn: Connection = await connect(PG_CONNECTION_URI)

    await conn.execute(
        "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
        "Bob",
        "secret_pw",
        datetime.date(1984, 3, 1),
    )

    await conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [
        CRUMBS_CONNECT,
        {
            "category": "query",
            "data": {
                "db.params": ["Bob", "secret_pw", "datetime.date(1984, 3, 1)"],
                "db.paramstyle": "format",
            },
            "message": "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
            "type": "default",
        },
    ]


@pytest.mark.asyncio
async def test_cursor(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn: Connection = await connect(PG_CONNECTION_URI)

    await conn.executemany(
        "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
        [
            ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
            ("Alice", "pw", datetime.date(1990, 12, 25)),
        ],
    )

    async with conn.transaction():
        # Postgres requires non-scrollable cursors to be created
        # and used in a transaction.
        async for record in conn.cursor(
            "SELECT * FROM users WHERE dob > $1", datetime.date(1970, 1, 1)
        ):
            print(record)

    await conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [
        CRUMBS_CONNECT,
        {
            "category": "query",
            "data": {"db.executemany": True},
            "message": "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
            "type": "default",
        },
        {"category": "query", "data": {}, "message": "BEGIN;", "type": "default"},
        {
            "category": "query",
            "data": {"db.cursor": mock.ANY},
            "message": "SELECT * FROM users WHERE dob > $1",
            "type": "default",
        },
        {"category": "query", "data": {}, "message": "COMMIT;", "type": "default"},
    ]


@pytest.mark.asyncio
async def test_cursor_manual(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn: Connection = await connect(PG_CONNECTION_URI)

    await conn.executemany(
        "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
        [
            ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
            ("Alice", "pw", datetime.date(1990, 12, 25)),
        ],
    )
    #
    async with conn.transaction():
        # Postgres requires non-scrollable cursors to be created
        # and used in a transaction.
        cur = await conn.cursor(
            "SELECT * FROM users WHERE dob > $1", datetime.date(1970, 1, 1)
        )
        record = await cur.fetchrow()
        print(record)
        while await cur.forward(1):
            record = await cur.fetchrow()
            print(record)

    await conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [
        CRUMBS_CONNECT,
        {
            "category": "query",
            "data": {"db.executemany": True},
            "message": "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
            "type": "default",
        },
        {"category": "query", "data": {}, "message": "BEGIN;", "type": "default"},
        {
            "category": "query",
            "data": {"db.cursor": mock.ANY},
            "message": "SELECT * FROM users WHERE dob > $1",
            "type": "default",
        },
        {
            "category": "query",
            "data": {"db.cursor": mock.ANY},
            "message": "SELECT * FROM users WHERE dob > $1",
            "type": "default",
        },
        {"category": "query", "data": {}, "message": "COMMIT;", "type": "default"},
    ]


@pytest.mark.asyncio
async def test_prepared_stmt(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    conn: Connection = await connect(PG_CONNECTION_URI)

    await conn.executemany(
        "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
        [
            ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
            ("Alice", "pw", datetime.date(1990, 12, 25)),
        ],
    )

    stmt = await conn.prepare("SELECT * FROM users WHERE name = $1")

    print(await stmt.fetchval("Bob"))
    print(await stmt.fetchval("Alice"))

    await conn.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [
        CRUMBS_CONNECT,
        {
            "category": "query",
            "data": {"db.executemany": True},
            "message": "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "SELECT * FROM users WHERE name = $1",
            "type": "default",
        },
    ]


@pytest.mark.asyncio
async def test_connection_pool(sentry_init, capture_events) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration()],
        _experiments={"record_sql_params": True},
    )
    events = capture_events()

    pool_size = 2

    pool = await asyncpg.create_pool(
        PG_CONNECTION_URI, min_size=pool_size, max_size=pool_size
    )

    async with pool.acquire() as conn:
        await conn.execute(
            "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
            "Bob",
            "secret_pw",
            datetime.date(1984, 3, 1),
        )

    async with pool.acquire() as conn:
        row = await conn.fetchrow("SELECT * FROM users WHERE name = $1", "Bob")
        assert row == (1, "Bob", "secret_pw", datetime.date(1984, 3, 1))

    await pool.close()

    capture_message("hi")

    (event,) = events

    for crumb in event["breadcrumbs"]["values"]:
        del crumb["timestamp"]

    assert event["breadcrumbs"]["values"] == [
        # The connection pool opens pool_size connections so we have the crumbs pool_size times
        *[CRUMBS_CONNECT] * pool_size,
        {
            "category": "query",
            "data": {},
            "message": "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "SELECT pg_advisory_unlock_all(); CLOSE ALL; UNLISTEN *; RESET ALL;",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "SELECT * FROM users WHERE name = $1",
            "type": "default",
        },
        {
            "category": "query",
            "data": {},
            "message": "SELECT pg_advisory_unlock_all(); CLOSE ALL; UNLISTEN *; RESET ALL;",
            "type": "default",
        },
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_disabled(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_options = {
        "integrations": [AsyncPGIntegration()],
        "traces_sample_rate": 1.0,
        "enable_db_query_source": False,
        "db_query_source_threshold_ms": 0,
        "_experiments": {
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    }

    sentry_init(**sentry_options)

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
            )

            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["name"] == "test_segment"
        assert insert_span["name"].startswith("INSERT INTO")
        assert connect_span["name"] == "connect"
        data = insert_span.get("attributes", {})
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
            )

            await conn.close()

        (event,) = events
        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")
        data = span.get("data", {})

    assert SPANDATA.CODE_LINENO not in data
    assert SPANDATA.CODE_NAMESPACE not in data
    assert SPANDATA.CODE_FILEPATH not in data
    assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.asyncio
@pytest.mark.parametrize("enable_db_query_source", [None, True])
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_enabled(
    sentry_init, capture_events, capture_items, enable_db_query_source, span_streaming
):
    sentry_options = {
        "integrations": [AsyncPGIntegration()],
        "traces_sample_rate": 1.0,
        "db_query_source_threshold_ms": 0,
        "_experiments": {
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    }
    if enable_db_query_source is not None:
        sentry_options["enable_db_query_source"] = enable_db_query_source

    sentry_init(**sentry_options)

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
            )

            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["name"] == "test_segment"
        assert insert_span["name"].startswith("INSERT INTO")
        assert connect_span["name"] == "connect"
        data = insert_span.get("attributes", {})
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
            )

            await conn.close()

        (event,) = events
        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")
        data = span.get("data", {})

    lineno_key = "code.line.number" if span_streaming else SPANDATA.CODE_LINENO
    filepath_key = "code.file.path" if span_streaming else SPANDATA.CODE_FILEPATH

    assert lineno_key in data
    assert filepath_key in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FUNCTION in data


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
            )

            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["name"] == "test_segment"
        assert insert_span["name"].startswith("INSERT INTO")
        assert connect_span["name"] == "connect"
        data = insert_span.get("attributes", {})
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
            )

            await conn.close()

        (event,) = events
        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")
        data = span.get("data", {})

    lineno_key = "code.line.number" if span_streaming else SPANDATA.CODE_LINENO
    filepath_key = "code.file.path" if span_streaming else SPANDATA.CODE_FILEPATH

    assert lineno_key in data
    assert filepath_key in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(lineno_key)) == int
    assert data.get(lineno_key) > 0
    assert (
        data.get(SPANDATA.CODE_NAMESPACE) == "tests.integrations.asyncpg.test_asyncpg"
    )
    assert data.get(filepath_key).endswith("tests/integrations/asyncpg/test_asyncpg.py")

    is_relative_path = data.get(filepath_key)[0] != os.sep
    assert is_relative_path

    assert data.get(SPANDATA.CODE_FUNCTION) == "test_query_source"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_with_module_in_search_path(
    sentry_init, capture_events, capture_items, span_streaming
):
    """
    Test that query source is relative to the path of the module it ran in
    """
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    from asyncpg_helpers.helpers import execute_query_in_connection

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await execute_query_in_connection(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                conn,
            )

            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["name"] == "test_segment"
        assert insert_span["name"].startswith("INSERT INTO")
        assert connect_span["name"] == "connect"
        data = insert_span.get("attributes", {})
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await execute_query_in_connection(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                conn,
            )

            await conn.close()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")
        data = span.get("data", {})

    lineno_key = "code.line.number" if span_streaming else SPANDATA.CODE_LINENO
    filepath_key = "code.file.path" if span_streaming else SPANDATA.CODE_FILEPATH

    assert lineno_key in data
    assert filepath_key in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(lineno_key)) == int
    assert data.get(lineno_key) > 0
    assert data.get(filepath_key) == "asyncpg_helpers/helpers.py"
    assert data.get(SPANDATA.CODE_NAMESPACE) == "asyncpg_helpers.helpers"

    is_relative_path = data.get(filepath_key)[0] != os.sep
    assert is_relative_path

    assert data.get(SPANDATA.CODE_FUNCTION) == "execute_query_in_connection"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_no_query_source_if_duration_too_short(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=100,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        @contextmanager
        def fake_record_sql_queries_streaming(*args, **kwargs):
            with record_sql_queries(*args, **kwargs) as span:
                pass
            span._start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            if span_streaming:
                span._end_timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)
            else:
                span._timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)
            yield span

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            with mock.patch(
                "sentry_sdk.integrations.asyncpg.record_sql_queries",
                fake_record_sql_queries_streaming,
            ):
                await conn.execute(
                    "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                )

            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        insert_span = spans[1]
        segment = spans[2]

        assert segment["name"] == "test_segment"
        assert insert_span["name"].startswith("INSERT INTO")
        assert connect_span["name"] == "connect"
        data = insert_span.get("attributes", {})
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)

            @contextmanager
            def fake_record_sql_queries(*args, **kwargs):
                with record_sql_queries(*args, **kwargs) as span:
                    pass
                span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
                span.timestamp = datetime.datetime(2024, 1, 1, microsecond=99999)
                yield span

            with mock.patch(
                "sentry_sdk.integrations.asyncpg.record_sql_queries",
                fake_record_sql_queries,
            ):
                await conn.execute(
                    "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
                )

            await conn.close()

        (event,) = events

        span = event["spans"][-1]
        assert span["description"].startswith("INSERT INTO")
        data = span.get("data", {})

    assert SPANDATA.CODE_LINENO not in data
    assert SPANDATA.CODE_NAMESPACE not in data
    assert SPANDATA.CODE_FILEPATH not in data
    assert SPANDATA.CODE_FUNCTION not in data


@pytest.mark.asyncio
async def test_query_source_if_duration_over_threshold(sentry_init, capture_events):
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=100,
    )

    events = capture_events()

    with start_transaction(name="test_transaction", sampled=True):
        conn: Connection = await connect(PG_CONNECTION_URI)

        @contextmanager
        def fake_record_sql_queries(*args, **kwargs):
            with record_sql_queries(*args, **kwargs) as span:
                pass
            span.start_timestamp = datetime.datetime(2024, 1, 1, microsecond=0)
            span.timestamp = datetime.datetime(2024, 1, 1, microsecond=100001)
            yield span

        with mock.patch(
            "sentry_sdk.integrations.asyncpg.record_sql_queries",
            fake_record_sql_queries,
        ):
            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES ('Alice', 'secret', '1990-12-25')",
            )

        await conn.close()

    (event,) = events

    span = event["spans"][-1]
    assert span["description"].startswith("INSERT INTO")

    data = span.get("data", {})

    assert SPANDATA.CODE_LINENO in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FILEPATH in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(SPANDATA.CODE_LINENO)) == int
    assert data.get(SPANDATA.CODE_LINENO) > 0
    assert (
        data.get(SPANDATA.CODE_NAMESPACE) == "tests.integrations.asyncpg.test_asyncpg"
    )
    assert data.get(SPANDATA.CODE_FILEPATH).endswith(
        "tests/integrations/asyncpg/test_asyncpg.py"
    )

    is_relative_path = data.get(SPANDATA.CODE_FILEPATH)[0] != os.sep
    assert is_relative_path

    assert (
        data.get(SPANDATA.CODE_FUNCTION)
        == "test_query_source_if_duration_over_threshold"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_span_origin(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.execute("SELECT 1")
            await conn.fetchrow("SELECT 2")
            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 4

        connect_span = spans[0]
        select1_span = spans[1]
        select2_span = spans[2]
        segment = spans[3]

        assert segment["name"] == "test_segment"
        assert connect_span["name"] == "connect"
        assert select1_span["name"] == "SELECT 1"
        assert select2_span["name"] == "SELECT 2"

        assert segment["attributes"]["sentry.origin"] == "manual"
        assert connect_span["attributes"]["sentry.origin"] == "auto.db.asyncpg"
        assert select1_span["attributes"]["sentry.origin"] == "auto.db.asyncpg"
        assert select2_span["attributes"]["sentry.origin"] == "auto.db.asyncpg"
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.execute("SELECT 1")
            await conn.fetchrow("SELECT 2")
            await conn.close()

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "manual"

        for span in event["spans"]:
            assert span["origin"] == "auto.db.asyncpg"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_multiline_query_description_normalized(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)
            await conn.execute(
                """
                SELECT
                    id,
                    name
                FROM
                    users
                WHERE
                    name = 'Alice'
                """
            )
            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 3

        connect_span = spans[0]
        select_span = spans[1]
        segment = spans[2]

        assert segment["name"] == "test_segment"
        assert connect_span["name"] == "connect"
        assert select_span["name"] == "SELECT id, name FROM users WHERE name = 'Alice'"
    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            conn: Connection = await connect(PG_CONNECTION_URI)
            await conn.execute(
                """
                SELECT
                    id,
                    name
                FROM
                    users
                WHERE
                    name = 'Alice'
                """
            )
            await conn.close()

        (event,) = events

        spans = [
            s
            for s in event["spans"]
            if s["op"] == "db" and "SELECT" in s.get("description", "")
        ]
        assert len(spans) == 1
        assert (
            spans[0]["description"] == "SELECT id, name FROM users WHERE name = 'Alice'"
        )


@pytest.mark.asyncio
async def test_before_send_transaction_sees_normalized_description(
    sentry_init, capture_events
):
    def before_send_transaction(event, hint):
        for span in event.get("spans", []):
            desc = span.get("description", "")
            if "SELECT id, name FROM users" in desc:
                span["description"] = "filtered"
        return event

    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        before_send_transaction=before_send_transaction,
    )
    events = capture_events()

    with start_transaction(name="test_transaction"):
        conn: Connection = await connect(PG_CONNECTION_URI)
        await conn.execute(
            """
            SELECT
                id,
                name
            FROM
                users
            """
        )
        await conn.close()

    (event,) = events
    spans = [
        s
        for s in event["spans"]
        if s["op"] == "db" and "filtered" in s.get("description", "")
    ]

    assert len(spans) == 1
    assert spans[0]["description"] == "filtered"


def _assert_query_source(span, span_streaming, expected_function):
    if span_streaming:
        data = span.get("attributes", {})
        lineno_key = "code.line.number"
        filepath_key = "code.file.path"
    else:
        data = span.get("data", {})
        lineno_key = SPANDATA.CODE_LINENO
        filepath_key = SPANDATA.CODE_FILEPATH

    assert lineno_key in data
    assert filepath_key in data
    assert SPANDATA.CODE_NAMESPACE in data
    assert SPANDATA.CODE_FUNCTION in data

    assert type(data.get(lineno_key)) == int
    assert data.get(lineno_key) > 0
    assert data[SPANDATA.CODE_NAMESPACE] == "tests.integrations.asyncpg.test_asyncpg"
    assert data.get(filepath_key).endswith("tests/integrations/asyncpg/test_asyncpg.py")
    assert data.get(filepath_key)[0] != os.sep
    assert data[SPANDATA.CODE_FUNCTION] == expected_function


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_execute(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)
            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
                "Alice",
                "pw",
                datetime.date(1990, 12, 25),
            )
            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]
        assert len(spans) == 3

        connect_span = spans[0]
        query_span = spans[1]
        segment = spans[2]

        assert connect_span["name"] == "connect"
        assert query_span["name"].startswith("INSERT INTO")
        assert segment["name"] == "test_segment"
        assert segment["is_segment"] is True
    else:
        events = capture_events()
        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)
            await conn.execute(
                "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
                "Alice",
                "pw",
                datetime.date(1990, 12, 25),
            )
            await conn.close()

        (event,) = events
        spans = event["spans"]
        assert len(spans) == 2
        assert spans[0]["description"] == "connect"
        assert spans[1]["description"].startswith("INSERT INTO")
        query_span = spans[1]

    _assert_query_source(query_span, span_streaming, "test_query_source_execute")


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_executemany(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)
            await conn.executemany(
                "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
                [("Bob", "secret_pw", datetime.date(1984, 3, 1))],
            )
            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]
        assert len(spans) == 3

        connect_span = spans[0]
        query_span = spans[1]
        segment = spans[2]

        assert connect_span["name"] == "connect"
        assert query_span["name"].startswith("INSERT INTO")
        assert segment["name"] == "test_segment"
    else:
        events = capture_events()
        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)
            await conn.executemany(
                "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
                [("Bob", "secret_pw", datetime.date(1984, 3, 1))],
            )
            await conn.close()

        (event,) = events
        spans = event["spans"]
        assert len(spans) == 2
        assert spans[0]["description"] == "connect"
        assert spans[1]["description"].startswith("INSERT INTO")
        query_span = spans[1]

    _assert_query_source(query_span, span_streaming, "test_query_source_executemany")


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_query_source_prepare(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)
            await conn.prepare("SELECT * FROM users WHERE name = $1")
            await conn.close()
        sentry_sdk.flush()

        spans = [item.payload for item in items]
        assert len(spans) == 3
        connect_span = spans[0]
        query_span = spans[1]
        segment = spans[2]

        assert connect_span["name"] == "connect"
        assert query_span["name"] == "SELECT * FROM users WHERE name = $1"
        assert segment["name"] == "test_segment"

        assert (
            query_span["attributes"][SPANDATA.DB_QUERY_TEXT]
            == "SELECT * FROM users WHERE name = $1"
        )
    else:
        events = capture_events()
        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)
            await conn.prepare("SELECT * FROM users WHERE name = $1")
            await conn.close()

        (event,) = events
        spans = event["spans"]
        assert len(spans) == 2
        assert spans[0]["description"] == "connect"
        assert spans[1]["description"] == "SELECT * FROM users WHERE name = $1"
        query_span = spans[1]

    _assert_query_source(query_span, span_streaming, "test_query_source_prepare")


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_cursor_iteration_creates_db_cursor_iter_spans(
    sentry_init, capture_events, capture_items, span_streaming
) -> None:
    """
    Regression test for https://github.com/getsentry/sentry-python/issues/6576

    When iterating a server-side cursor with a small prefetch, asyncpg fetches
    rows in batches. Each batch triggers BaseCursor._bind_exec (on first query) and
    BaseCursor._exec (second query onwards) through CursorIterator.__anext__, which creates a
    span with the same query description. The resulting burst of identical spans
    causes Sentry's N+1 query detector to raise a false positive.

    To mitigate, we set the "op"/"sentry.op" to `db.cursor.iter` instead of `db`
    so that the sentry backend can exclude these spans from n+1 detection.
    """
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.executemany(
                "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
                [(f"user-{i}", "pw", datetime.date(1990, 1, 1)) for i in range(20)],
            )

            async with conn.transaction():
                async for _record in conn.cursor("SELECT * FROM users", prefetch=5):
                    pass

            await conn.close()

        sentry_sdk.flush()

        cursor_iter_spans = [
            item.payload
            for item in items
            if item.payload.get("name") == "SELECT * FROM users"
        ]

        assert len(cursor_iter_spans) == 5
        for span in cursor_iter_spans:
            assert span["attributes"]["sentry.op"] == OP.DB_CURSOR_ITERATOR
            assert span["attributes"][SPANDATA.DB_QUERY_TEXT] == "SELECT * FROM users"
    else:
        events = capture_events()

        with start_transaction(name="test_transaction", sampled=True):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.executemany(
                "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
                [(f"user-{i}", "pw", datetime.date(1990, 1, 1)) for i in range(20)],
            )

            async with conn.transaction():
                async for _record in conn.cursor("SELECT * FROM users", prefetch=5):
                    pass

            await conn.close()

        (event,) = events

        cursor_iter_spans = [
            s for s in event["spans"] if s.get("description") == "SELECT * FROM users"
        ]

        assert len(cursor_iter_spans) == 5
        for span in cursor_iter_spans:
            assert span["op"] == OP.DB_CURSOR_ITERATOR


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_cursor_fetch_methods_create_spans(
    sentry_init, capture_events, capture_items, span_streaming
) -> None:
    sentry_init(
        integrations=[AsyncPGIntegration()],
        traces_sample_rate=1.0,
        enable_db_query_source=True,
        db_query_source_threshold_ms=0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items()

        with sentry_sdk.traces.start_span(name="test_segment"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.executemany(
                "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
                [
                    ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
                    ("Alice", "pw", datetime.date(1990, 12, 25)),
                ],
            )

            async with conn.transaction():
                cur = await conn.cursor(
                    "SELECT * FROM users WHERE dob > $1", datetime.date(1970, 1, 1)
                )
                # These exercise the `_exec` patch
                await cur.fetchrow()
                await cur.fetchrow()

            await conn.close()

        sentry_sdk.flush()

        spans = [item.payload for item in items]

        assert len(spans) == 7

        connect_span = spans[0]
        executemany_span = spans[1]
        begin_span = spans[2]
        fetchrow_span_1 = spans[3]
        fetchrow_span_2 = spans[4]
        commit_span = spans[5]
        _segment_span = spans[6]

        assert connect_span["name"] == "connect"
        assert (
            executemany_span["name"]
            == "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)"
        )
        assert begin_span["name"] == "BEGIN;"
        assert fetchrow_span_1["name"] == "SELECT * FROM users WHERE dob > $1"
        assert fetchrow_span_2["name"] == "SELECT * FROM users WHERE dob > $1"
        assert commit_span["name"] == "COMMIT;"

        assert (
            fetchrow_span_1["attributes"][SPANDATA.DB_QUERY_TEXT]
            == "SELECT * FROM users WHERE dob > $1"
        )
        assert (
            fetchrow_span_2["attributes"][SPANDATA.DB_QUERY_TEXT]
            == "SELECT * FROM users WHERE dob > $1"
        )

        for span in (fetchrow_span_1, fetchrow_span_2):
            assert span["attributes"][SPANDATA.DB_SYSTEM_NAME] == "postgresql"
            assert span["attributes"][SPANDATA.DB_DRIVER_NAME] == "asyncpg"
            assert span["attributes"]["sentry.op"] == OP.DB_CURSOR_FETCH
            assert span["attributes"]["sentry.origin"] == "auto.db.asyncpg"

    else:
        events = capture_events()

        with start_transaction(name="test_transaction"):
            conn: Connection = await connect(PG_CONNECTION_URI)

            await conn.executemany(
                "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)",
                [
                    ("Bob", "secret_pw", datetime.date(1984, 3, 1)),
                    ("Alice", "pw", datetime.date(1990, 12, 25)),
                ],
            )

            async with conn.transaction():
                cur = await conn.cursor(
                    "SELECT * FROM users WHERE dob > $1", datetime.date(1970, 1, 1)
                )
                # These exercise the `_exec` patch
                await cur.fetchrow()
                await cur.fetchrow()

            await conn.close()

        (event,) = events

        assert len(event["spans"]) == 6

        connect_span = event["spans"][0]
        executemany_span = event["spans"][1]
        begin_span = event["spans"][2]
        fetchrow_span_1 = event["spans"][3]
        fetchrow_span_2 = event["spans"][4]
        commit_span = event["spans"][5]

        assert connect_span["description"] == "connect"
        assert (
            executemany_span["description"]
            == "INSERT INTO users(name, password, dob) VALUES($1, $2, $3)"
        )
        assert begin_span["description"] == "BEGIN;"
        assert fetchrow_span_1["description"] == "SELECT * FROM users WHERE dob > $1"
        assert fetchrow_span_2["description"] == "SELECT * FROM users WHERE dob > $1"
        assert commit_span["description"] == "COMMIT;"

        for span in (fetchrow_span_1, fetchrow_span_2):
            assert span["data"]["db.cursor"] is not None
            assert span["data"]["db.system"] == "postgresql"
            assert span["data"]["db.driver.name"] == "asyncpg"
            assert span["op"] == OP.DB_CURSOR_FETCH
            assert span["origin"] == "auto.db.asyncpg"

    _assert_query_source(
        span,
        span_streaming,
        "test_cursor_fetch_methods_create_spans",
    )



import pytest

import sentry_sdk
from sentry_sdk import start_span


def test_conversation_id_propagates_to_span_with_gen_ai_operation_name(
    sentry_init, capture_events
):
    """Span with gen_ai.operation.name data should get conversation_id."""
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    scope = sentry_sdk.get_current_scope()
    scope.set_conversation_id("conv-op-name-test")

    with sentry_sdk.start_transaction(name="test-tx"):
        with start_span(op="http.client") as span:
            span.set_data("gen_ai.operation.name", "chat")

    (event,) = events
    span_data = event["spans"][0]["data"]
    assert span_data.get("gen_ai.conversation.id") == "conv-op-name-test"


def test_conversation_id_propagates_to_span_with_ai_op(sentry_init, capture_events):
    """Span with ai.* op should get conversation_id."""
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    scope = sentry_sdk.get_current_scope()
    scope.set_conversation_id("conv-ai-op-test")

    with sentry_sdk.start_transaction(name="test-tx"):
        with start_span(op="ai.chat.completions"):
            pass

    (event,) = events
    span_data = event["spans"][0]["data"]
    assert span_data.get("gen_ai.conversation.id") == "conv-ai-op-test"


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
def test_conversation_id_propagates_to_span_with_gen_ai_op(
    sentry_init, capture_events, capture_items, stream_gen_ai_spans
):
    """Span with gen_ai.* op should get conversation_id."""
    sentry_init(
        traces_sample_rate=1.0,
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    if stream_gen_ai_spans:
        items = capture_items("span")

        scope = sentry_sdk.get_current_scope()
        scope.set_conversation_id("conv-gen-ai-op-test")

        with sentry_sdk.start_transaction(name="test-tx"):
            with start_span(op="gen_ai.invoke_agent"):
                pass

        spans = [item.payload for item in items if item.type == "span"]
        span_data = spans[0]["attributes"]
    else:
        events = capture_events()

        scope = sentry_sdk.get_current_scope()
        scope.set_conversation_id("conv-gen-ai-op-test")

        with sentry_sdk.start_transaction(name="test-tx"):
            with start_span(op="gen_ai.invoke_agent"):
                pass

        (event,) = events
        span_data = event["spans"][0]["data"]

    assert span_data.get("gen_ai.conversation.id") == "conv-gen-ai-op-test"


def test_conversation_id_not_propagated_to_non_ai_span(sentry_init, capture_events):
    """Non-AI span should NOT get conversation_id."""
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    scope = sentry_sdk.get_current_scope()
    scope.set_conversation_id("conv-should-not-appear")

    with sentry_sdk.start_transaction(name="test-tx"):
        with start_span(op="http.client") as span:
            span.set_data("some.other.data", "value")

    (event,) = events
    span_data = event["spans"][0]["data"]
    assert "gen_ai.conversation.id" not in span_data


def test_conversation_id_not_propagated_when_not_set(sentry_init, capture_events):
    """AI span should not have conversation_id if not set on scope."""
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    # Ensure no conversation_id is set
    scope = sentry_sdk.get_current_scope()
    scope.remove_conversation_id()

    with sentry_sdk.start_transaction(name="test-tx"):
        with start_span(op="ai.chat.completions"):
            pass

    (event,) = events
    span_data = event["spans"][0]["data"]
    assert "gen_ai.conversation.id" not in span_data


def test_conversation_id_not_propagated_to_span_without_op(sentry_init, capture_events):
    """Span without op and without gen_ai.operation.name should NOT get conversation_id."""
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    scope = sentry_sdk.get_current_scope()
    scope.set_conversation_id("conv-no-op-test")

    with sentry_sdk.start_transaction(name="test-tx"):
        with start_span(name="unnamed-span") as span:
            span.set_data("regular.data", "value")

    (event,) = events
    span_data = event["spans"][0]["data"]
    assert "gen_ai.conversation.id" not in span_data


def test_conversation_id_propagates_with_gen_ai_operation_name_no_op(
    sentry_init, capture_events
):
    """Span with gen_ai.operation.name but no op should still get conversation_id."""
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    scope = sentry_sdk.get_current_scope()
    scope.set_conversation_id("conv-no-op-but-data-test")

    with sentry_sdk.start_transaction(name="test-tx"):
        with start_span(name="unnamed-span") as span:
            span.set_data("gen_ai.operation.name", "embedding")

    (event,) = events
    span_data = event["spans"][0]["data"]
    assert span_data.get("gen_ai.conversation.id") == "conv-no-op-but-data-test"


def test_conversation_id_propagates_to_transaction_with_ai_op(
    sentry_init, capture_events
):
    """Transaction with ai.* op should get conversation_id."""
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    scope = sentry_sdk.get_current_scope()
    scope.set_conversation_id("conv-tx-ai-op-test")

    with sentry_sdk.start_transaction(op="ai.workflow", name="AI Workflow"):
        pass

    (event,) = events
    trace_data = event["contexts"]["trace"]["data"]
    assert trace_data.get("gen_ai.conversation.id") == "conv-tx-ai-op-test"


def test_conversation_id_not_propagated_to_non_ai_transaction(
    sentry_init, capture_events
):
    """Non-AI transaction should NOT get conversation_id."""
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    scope = sentry_sdk.get_current_scope()
    scope.set_conversation_id("conv-tx-should-not-appear")

    with sentry_sdk.start_transaction(op="http.server", name="HTTP Request"):
        pass

    (event,) = events
    trace_data = event["contexts"]["trace"]["data"]
    assert "gen_ai.conversation.id" not in trace_data
